# -*- coding: utf-8 -*-

#changement de dossier
import os
os.chdir("D:/DataMining/Databases_for_mining/dataset_for_soft_dev_and_comparison/discriminant_analysis/python")

#importation de l'échantillon train
import pandas
DTrain = pandas.read_excel("Eau_de_vie_LDA_python.xlsx",sheet_name="TRAIN")
print(DTrain.info())

#préparation des structures
yTrain = DTrain.iloc[:,0]
XTrain = DTrain.iloc[:,1:]

#effectif -- 52
n = DTrain.shape[0]

#nombre de descripteurs -- 8
p = XTrain.shape[1]

#nombre de classes -- 3
K = len(yTrain.value_counts())

#distribution des classes
print(yTrain.value_counts(normalize=True))

#version
import sklearn
print(sklearn.__version__)

#importation
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis

#instanciation
lda = LinearDiscriminantAnalysis(solver="eigen",store_covariance=True)

#apprentissage
lda.fit(XTrain,yTrain)

#liste des champs
print(dir(lda))

#affichage brut des coefficients
print(lda.coef_)

#dimensions
print(lda.coef_.shape)

#liste des modalités
print(lda.classes_)

#structure temporaire pour affichage des coefficients
tmp= pandas.DataFrame(lda.coef_.transpose(),columns=lda.classes_,index=XTrain.columns)
print(tmp)

#et la constante
print(lda.intercept_)

#modifier le mode d'affichage des matrices numpy
#pas de notation scientifique
import numpy
numpy.set_printoptions(suppress=True,linewidth=120)

#covariance
print(lda.covariance_)

#pour retrouver la matrice W de SAS
W = n/(n-K)*lda.covariance_
print(W)

#calculons son déterminant
detW = numpy.linalg.det(W)
print(detW)

#et son logarithme
logDetW = numpy.log(detW)
print(logDetW)

#covariance totale
T = numpy.cov(XTrain.values,rowvar=False)
print(T)

#*****************************
#évaluation globale
#*****************************

#lambda de Wilks
LW = numpy.linalg.det(lda.covariance_)/numpy.linalg.det((n-1)/n*T)
print(LW)

#F de Rao

#ddl numérateur
ddlNum = p * (K-1)
print(ddlNum)

#valeur intermédiaire
temp = p**2 + (K-1)**2 - 5
temp = numpy.where(temp > 0,numpy.sqrt(((p**2) * ((K-1)**2) - 4)/temp),1)

#ddl dénominateur
ddlDenom = (2*n-p-K-2)/2 * temp - (ddlNum - 2)/2
print(ddlDenom)

#stat de test
FRao = LW**(1/temp)
FRao = ((1-FRao)/FRao)*(ddlDenom/ddlNum)
print(FRao)

#p-value
import scipy
print(1-scipy.stats.f.cdf(FRao,ddlNum,ddlDenom))

#*****************************
#évaluation de chaque variable
#*****************************

#degrés de liberté - numérateur
ddlSuppNum = K-1

#degré de liberté dénominateur
ddlSuppDenom = n - K - p + 1

#FTest - préparation
FTest = numpy.zeros(p)

#p-value pour FTest
pvalueFTest = numpy.zeros(p)

#Tester chaque variable
for j in range(p):
    #matricesintermédiaires numérateur
    tempNum = lda.covariance_.copy()
    #supprimer la référence de la variable à traiter
    tempNum = numpy.delete(tempNum,j,axis=0)
    tempNum = numpy.delete(tempNum,j,axis=1)
    #même chose pour dénominateur
    tempDenom = (n-1)/n*T
    tempDenom = numpy.delete(tempDenom,j,axis=0)
    tempDenom = numpy.delete(tempDenom,j,axis=1)
    #lambda sans la variable
    LWVar = numpy.linalg.det(tempNum)/numpy.linalg.det(tempDenom)
    print(LWVar)
    #FValue
    FValue = ddlSuppDenom/ddlSuppNum*(LWVar/LW-1)
    #récupération des valeurs
    FTest[j] = FValue
    pvalueFTest[j] = 1 - scipy.stats.f.cdf(FValue,ddlSuppNum,ddlSuppDenom)
    
#affichage
temp = {'var':XTrain.columns,'F':FTest,'pvalue':pvalueFTest}
print(pandas.DataFrame(temp))

#*****************************
#évaluation en test
#*****************************

#chargement échantillon test
DTest = pandas.read_excel("Eau_de_vie_LDA_python.xlsx",sheet_name="TEST")
print(DTest.info())

#récupérer la matrice des X
XTest = DTest.iloc[:,1:]

#distribution des classes
print(DTest.TYPE.value_counts(normalize=True))

#la fonction "score" fournit le taux de succès
print(lda.score(XTest,DTest.TYPE))

#prediction simple
pred = lda.predict(XTest)

#distribution des classes prédites
print(numpy.unique(pred,return_counts=True))

#matrice de confusion
from sklearn import metrics
print(metrics.confusion_matrix(DTest.TYPE,pred))                   

#rapport
print(metrics.classification_report(DTest.TYPE,pred))


#probabilités d'affectation des 15 premiers
proba = lda.predict_proba(XTest)
print(proba[:15,:])

#affectation pour individu n°14
print(pred[14])

#proba individu n°14
print(proba[14,:])

#individu n°14
print(DTest.iloc[14,:])

#application de la fonction de classement
#KIRSCH
fKIRSCH = numpy.sum(DTest.iloc[14,1:] * lda.coef_[0,:]) + lda.intercept_[0]
print(fKIRSCH)

#MIRAB
fMIRAB = numpy.sum(DTest.iloc[14,1:] * lda.coef_[1,:]) + lda.intercept_[1]
print(fMIRAB)

#POIRE
fPOIRE = numpy.sum(DTest.iloc[14,1:] * lda.coef_[2,:]) + lda.intercept_[2]
print(fPOIRE)

#dénominateur
denomSoftMax = numpy.exp(fKIRSCH)+numpy.exp(fMIRAB)+numpy.exp(fPOIRE)

#softmax KIRSCH
pKIRSCH = numpy.exp(fKIRSCH)/denomSoftMax
print(pKIRSCH)

#softmax MIRAB
pMIRAB = numpy.exp(fMIRAB)/denomSoftMax
print(pMIRAB)

#softmax POIRE
pPOIRE = numpy.exp(fPOIRE)/denomSoftMax
print(pPOIRE)

#affichage des probas d'affectation estimées
print("(KIRSCH : %.8f ; MIRAB : %.8f : POIRE : %.8f)" % (pKIRSCH,pMIRAB,pPOIRE))

#*********************************
#analyse factorielle discriminante
#*********************************

#projection dans l'espace facoriel
Fact = lda.transform(XTrain)
print(Fact[:5,:])

#data frame pour représentation graphique
dfFact = pandas.DataFrame(Fact,columns=['F1','F2'])
dfFact['classe'] = DTrain.TYPE

#représentation graphique
import seaborn as sns
sns.scatterplot(x="F1",y="F2",hue="classe",data=dfFact)

#************************
#** dessin des frontières
#************************

#chargement des données
DTom = pandas.read_excel("Tomassone_et_al.xlsx",sheet_name=0)
print(DTom.info())

#graphique
sns.scatterplot(x="X1",y="X2",hue="Groupe",data=DTom)

#lda
ldaTom = LinearDiscriminantAnalysis(solver="eigen")

#apprentissage
ldaTom.fit(DTom.iloc[:,1:],DTom.Groupe)

#fonction de classement
print(ldaTom.coef_)

#constante
print(ldaTom.intercept_)

#valeurs de x1
x1 = numpy.arange(0,14,step=1)

#coordonnées x2
x2 = (-ldaTom.intercept_ - ldaTom.coef_[0][0] * x1) / ldaTom.coef_[0][1]

#data frame
dfTemp = pandas.DataFrame({'x1':x1,'x2':x2})

#graphique des points
ax = sns.scatterplot(x="X1",y="X2",hue="Groupe",data=DTom)
#ajouter les numéros d'individus
for i in DTom.index:
    ax.text(DTom.X1[i]+0.2,DTom.X2[i],str(DTom.index[i]))
#rajouter la frontière de séparation
sns.lineplot(x='x1',y='x2',data=dfTemp)

#************************
#** dessin des frontières
#************************

#distance à la frontière
dist = ldaTom.decision_function(DTom.iloc[:,1:])
print(pandas.DataFrame(dist,index=DTom.index))

#vérifions pour les points:
# => la distance n'est pas normée
# => et elle est signée
d = ldaTom.coef_[0][0]*DTom.X1+ldaTom.coef_[0][1]*DTom.X2+ldaTom.intercept_
print(d)

#et la probabilité d'affectation est proportionnelle à cette distance
print(pandas.DataFrame(ldaTom.predict_proba(DTom.iloc[:,1:]),index=DTom.index))
